
#ifndef prtrows
#  define prtrows 10
#endif 
#ifndef prtcols
#  define prtcols 10
#endif 

module heat_mod
 
  ! precision parameter for 64bit numbers
  INTEGER, PARAMETER :: DP = 8
  integer :: m, n, itmax 
 
  REAL (KIND=DP), dimension (:,:), allocatable :: phi [:,:]
  REAL (KIND=DP), dimension (:,:), allocatable :: phin

  integer :: numprocs, myrank, mprocs,nprocs, iproc,kproc 
 
 contains 
  subroutine heatpr
  
    implicit none
  
    integer :: i_global, k_global, ip, kp
    integer :: i, k, prt_i,prt_k, prt_i_max, prt_k_max
    integer :: klist(0:19) 
    REAL (KIND=DP) :: vallist(0:19) 

    ! Each process handles only local i=1..m-2, k=1..n-2.
    ! Boundary conditions are at i=0, k=0, i=m-1, k=n-1
    ! on first and last processes. The total range
    ! is therefore i_global=0..mprocs*(m-2)+1
    ! and k_global=0..nprocs*(n-2)+1. */
 
   sync all 
   if (myrank == 0) then 
    prt_i_max = min(prtrows, mprocs*(m-2)+2) 
    prt_k_max = min(prtcols, nprocs*(n-2)+2, 20) 
  
    do prt_k = 0, prt_k_max-1
       k_global=1.0*prt_k/(prt_k_max-1)*(nprocs*(n-2)+1) + 0.5  ! rounding 
       klist(prt_k) = k_global
    end do
  
    write (*,'(''i    \k='',20(''  '',I6))') (klist(prt_k), prt_k = 0,prt_k_max-1)
    write (*,'(''------\-'',20A)') ('--------', prt_k = 0,prt_k_max-1)
  
    do prt_i = 0, prt_i_max-1
       i_global=1.0*prt_i/(prt_i_max-1)*(mprocs*(m-2)+1) + 0.5  ! rounding
       if (i_global == 0) then
         ip=0
         i=0
       else if (i_global == mprocs*(m-2)+1) then
         ip=mprocs-1
         i=m-1
       else  
         ip = (i_global-1)/(m-2)
         i = mod((i_global-1),(m-2)) + 1
       end if 
       do prt_k = 0, prt_k_max-1
         k_global=klist(prt_k)
         if (k_global == 0) then
           kp=0
           k=0
         else if (k_global == nprocs*(n-2)+1) then
           kp=nprocs-1
           k=n-1
         else  
           kp = (k_global-1)/(n-2)
           k = mod((k_global-1),(n-2)) + 1
         end if 
         vallist(prt_k) = phi(i,k)[ip,kp]
       end do 
       write (*, '(I6,'' |'',20(2X,F6.4))') &
            i_global, (vallist(prt_k), prt_k = 0,prt_k_max-1) 
    end do
   end if ! myrank==0
   sync all 
  
    return
  
  end subroutine heatpr


  subroutine halo_exchange
    implicit none
    sync all
    if (iproc>0) then
      phi(0,1:n-2) = phi(m-2,1:n-2)[iproc-1,kproc] 
    end if
    if (iproc<mprocs-1) then
      phi(m-1,1:n-2) = phi(1,1:n-2)[iproc+1,kproc] 
    end if
    if (kproc>0) then
      phi(1:m-2,0) = phi(1:m-2,n-2)[iproc,kproc-1] 
    end if
    if (kproc<nprocs-1) then
      phi(1:m-2,n-1) = phi(1:m-2,1)[iproc,kproc+1] 
    end if
    sync all
  end subroutine halo_exchange

  subroutine w_time(wt) 
    implicit none 
    REAL (KIND=DP) :: wt
    character(len=10) :: timestring
    character(len=8)  :: datestring
    integer hh, mm 
    call date_and_time(datestring,timestring)
    read(timestring,'(i2,i2,f6.3)') hh, mm, wt
    wt = wt + 60d0*mm + 3600d0*hh
    return
  end subroutine w_time
 
  subroutine heat_sub
  
    implicit none 
  
    integer :: i,k,it 
  
    REAL (KIND=DP), parameter :: eps = 1.0d-08

    ! the following arrays are globally defined in the module header 
    ! and are allocated in the main program with the following index ranges
    ! REAL (KIND=DP), dimension (0:m-1,0:n-1) :: phi[0:mprocs-1,0:nprocs-1]
    ! REAL (KIND=DP), dimension (1:m-2,1:n-2) :: phin

    ! for CAF parallelization of the collective maximum-reduction of dphimax variable 
    REAL (KIND=DP), SAVE :: mydphimax[0:*] 
    integer :: rank 
  
    REAL (KIND=DP) :: dx,dy,dx2,dy2,dx2i,dy2i,dt,dphi,dphimax
  
    ! times using cpu_time
    real :: t0, t1
    REAL (KIND=DP) :: wt0, wt1
  
    ! END DECLARATION
  
    dx = 1.0_DP / ((n-2)*nprocs+1)
    dy = 1.0_DP / ((m-2)*mprocs+1)
    dx2 = dx*dx
    dy2 = dy*dy
    dx2i = 1.0_DP / dx2
    dy2i = 1.0_DP / dy2
    dt = min(dx2,dy2) / 4.0_DP
  
    ! start values - total area: 0.d0
    do k = 0, n-1
       do i = 0, m-1
          phi(i,k) = 0.0_DP
       enddo
    enddo
  
    ! start values - boundary: 1.d0
   if (kproc == nprocs-1) then  ! CAF-distributed: only if highest kproc: 
    do i=0, m-1
       phi(i,n-1) = 1.0_DP
    enddo
   end if 
  
    ! start values - boundary: 0.d0
   if (kproc == 0) then  ! CAF-distributed: only if lowest kproc: 
    do i=0, m-1
       phi(i,0) = 0.0_DP
    enddo
   end if 
  
    ! start values - boundary: continuously from 0 to 1  
   if (iproc == 0) then  ! CAF-distributed: only if lowest iproc: 
    do k = 1,n-2
       phi(0,k)   = (k+kproc*(n-2))*dx
    enddo
   end if 
  
    ! start values - boundary: continuously from 0 to 1  
   if (iproc == mprocs-1) then  ! CAF-distributed: only if highest iproc: 
    do k = 1,n-2
       phi(m-1,k) = (k+kproc*(n-2))*dx
    enddo
   end if 
  
   if (myrank == 0) then 
    ! print array
    write (*,'(/,a)') 'Heat Conduction 2d'
    write (*,'(/,4(a,1pg12.4))') 'dx =',dx,', dy =',dy,', dt =',dt,', eps =',eps
    write (*,'(/,a)') 'start values'
   end if 
  
    call heatpr

    call w_time(wt0)
    call cpu_time(t0)
  
    ! iteration
    do it = 1,itmax
       call halo_exchange  
       dphimax = 0.0
  
       do k = 1,n-2
          do i = 1,m-2
             dphi = (phi(i+1,k)+phi(i-1,k)-2.*phi(i,k))*dy2i &
                  +(phi(i,k+1)+phi(i,k-1)-2.*phi(i,k))*dx2i
             dphi = dphi*dt
             dphimax = max(dphimax,dphi)
             phin(i,k) = phi(i,k)+dphi
          enddo
       enddo
  
       ! save values
       do k = 1,n-2
          do i =1,m-2
             phi(i,k) =phin(i,k)
          enddo
       enddo
  
       ! collective maximum-reduction of dphimax values:
       sync all  ! to guarantee that mydphimax[0] is read on all threads in previous "it" loop iteration
       mydphimax[myrank]=dphimax;
       sync all  ! to guarantee that all mydphimax[] arre stored and now readable for thread 0
       if (myrank == 0) then
         dphimax=0.0;
         do rank=0, numprocs-1 
           dphimax = max(dphimax, mydphimax[rank])
         end do
         mydphimax[0] = dphimax 
       end if
       sync all  ! to guarantee that mydphimax[0] is updated and readable for all threads
       dphimax = mydphimax[0]; 
 
       if(dphimax < eps) EXIT
  
    enddo
    if (dphimax >= eps) it=it-1
  
    call cpu_time(t1)
    call w_time(wt1)
    if (wt1<wt0) wt1 = wt1 + 24*3600d0 
 
   if (myrank == 0) then 
    ! print array
    write (*,'(/,a,i6,a)') 'phi after',it,' iterations'
   end if 
  
    call heatpr
  
   if (myrank == 0) then 
    write (*,'(/,a,f12.6)')  'cpu_time  :   ', t1-t0
    write (*,'(  a, f9.3)')  'wallclock :   ', wt1-wt0
   end if 
  
  
  end subroutine heat_sub
 
end module heat_mod
 
 
program heat

  use heat_mod 
  implicit none
 
  integer :: dim1, dim2, argc, arglen
  character(len=20) :: argval
 
  numprocs = num_images()
  myrank   = this_image()-1

  argc = command_argument_count()
  if (argc < 2) then
    if (myrank == 0) then
      print *,'Synopsis: heat <dim1> <dim2> [<itmax> [<mprocs>]]'
      print *
      print *,'Examples: heat 30 30'
      print *,'          heat 30 30 15000'
      print *,'          heat 30 30 15000 2'
    end if 
    stop
  end if 
  call get_command_argument(1, argval, arglen)
  read(argval,*) dim1
  call get_command_argument(2, argval, arglen)
  read(argval,*) dim2
  itmax=15000 
  if (argc >= 3) then
    call get_command_argument(3, argval, arglen)
    read(argval,*) itmax
  end if 
  if (argc >= 4) then
    call get_command_argument(4, argval, arglen)
    read(argval,*) mprocs
  else
    !  Factorization of numprocs:
    !    Goals: mprocs * nprocs = numprocs 
    !           dim1 / dim2 ~= mprocs / nprocs
    !    This would imply: mprocs ~= sqrt(numprocs*dim1/dim2) 
    mprocs = sqrt((numprocs+0.0)*dim1/dim2) + 0.5  ! rounded
  end if 
  if (mprocs < 1) mprocs=1
  do while (mprocs > 1)
    if (numprocs/mprocs*mprocs == numprocs) exit
    mprocs=mprocs-1
  end do
  nprocs = numprocs/mprocs;
 
  ! Process ranks:
  iproc = mod(myrank,mprocs)
  kproc = myrank/mprocs  
  
  ! Calculation of local dim1 x dim2
  dim1 = (dim1-1)/mprocs + 1  ! rounded up 
  dim2 = (dim2-1)/nprocs + 1  ! rounded up 
 
  m=dim1+2  ! two additional elements for the boundary conditions 
  n=dim2+2
 if (myrank == 0) then
  print '(''main:   dim1  ='',i4,''   dim2  ='',i4,''   itmax=''i6)', dim1, dim2, itmax
  print '(''main: m=dim1+2='',i4,'' n=dim2+2='',i4)', m, n
  print '(''main:  mprocs='',i4,'',    nprocs='',i4,'',  THREADS='',i4)', mprocs, nprocs, numprocs
 end if 
  
! allocate( phi(0:m-1,0:n-1)[0:mprocs-1,0:nprocs-1] )  ! the last value must be *
  allocate( phi(0:m-1,0:n-1)[0:mprocs-1,0:   *    ] ) 
  allocate( phin(1:m-2,1:n-2) )

  call heat_sub
 
end program heat
