
#ifndef prtrows
#  define prtrows 10
#endif 
#ifndef prtcols
#  define prtcols 10
#endif 

module heat_mod
 
  ! precision parameter for 64bit numbers
  INTEGER, PARAMETER :: DP = 8
  integer :: m, n, itmax 
 
  REAL (KIND=DP), dimension (:,:), allocatable :: phi
  REAL (KIND=DP), dimension (:,:), allocatable :: phin
 
 contains 
  subroutine heatpr
  
    implicit none
  
    integer :: i,k, prt_i,prt_k, prt_i_max, prt_k_max
    integer :: klist(0:19) 
 
    prt_i_max = min(prtrows, m) 
    prt_k_max = min(prtcols, n, 20) 
  
    do prt_k = 0, prt_k_max-1
       k = 1.0*prt_k/(prt_k_max-1)*(n-1) + 0.5   ! rounding
       klist(prt_k) = k
    end do
  
    write (*,'(''i    \k='',20(''  '',I6))') (klist(prt_k), prt_k = 0,prt_k_max-1)
    write (*,'(''------\-'',20A)') ('--------', prt_k = 0,prt_k_max-1)
  
    do prt_i = 0, prt_i_max-1
       i = 1.0*prt_i/(prt_i_max-1)*(m-1) + 0.5   ! rounding
       write (*, '(I6,'' |'',20(2X,F6.4))') &
            i, (phi(i,klist(prt_k)), prt_k = 0,prt_k_max-1) 
    end do
  
  
    return
  
  end subroutine heatpr

  subroutine w_time(wt) 
    implicit none 
    REAL (KIND=DP) :: wt
    character(len=10) :: timestring
    character(len=8)  :: datestring
    integer hh, mm 
    call date_and_time(datestring,timestring)
    read(timestring,'(i2,i2,f6.3)') hh, mm, wt
    wt = wt + 60d0*mm + 3600d0*hh
    return
  end subroutine w_time
 
  subroutine heat_sub
  
    implicit none 
  
    integer :: i,k,it 
  
    REAL (KIND=DP), parameter :: eps = 1.0d-08

    ! the following arrays are globally defined in the module header 
    ! and are allocated in the main program with the following index ranges
    ! REAL (KIND=DP), dimension (0:m-1,0:n-1) :: phi
    ! REAL (KIND=DP), dimension (1:m-2,1:n-2) :: phin
  
    REAL (KIND=DP) :: dx,dy,dx2,dy2,dx2i,dy2i,dt,dphi,dphimax
  
    ! times using cpu_time
    real :: t0, t1
    REAL (KIND=DP) :: wt0, wt1
  
    ! END DECLARATION
  
    dx = 1.0_DP / (n-1)
    dy = 1.0_DP / (m-1)
    dx2 = dx*dx
    dy2 = dy*dy
    dx2i = 1.0_DP / dx2
    dy2i = 1.0_DP / dy2
    dt = min(dx2,dy2) / 4.0_DP
  
    ! start values - total area: 0.d0
    do k = 0, n-1
       do i = 0, m-1
          phi(i,k) = 0.0_DP
       enddo
    enddo
  
    ! start values - boundary: 1.d0
    do i=0, m-1
       phi(i,n-1) = 1.0_DP
    enddo
  
    ! start values - boundary: 0.d0
    do i=0, m-1
       phi(i,0) = 0.0_DP
    enddo
  
    ! start values - boundary: continuously from 0 to 1  
    do k = 1,n-2
       phi(0,k)   = k*dx
       phi(m-1,k) = k*dx
    enddo
  
    ! print array
    write (*,'(/,a)') 'Heat Conduction 2d'
    write (*,'(/,4(a,1pg12.4))') 'dx =',dx,', dy =',dy,', dt =',dt,', eps =',eps
    write (*,'(/,a)') 'start values'
  
    call heatpr
 
    call w_time(wt0)
    call cpu_time(t0)
  
    ! iteration
    do it = 1,itmax
       dphimax = 0.0
  
       do k = 1,n-2
          do i = 1,m-2
             dphi = (phi(i+1,k)+phi(i-1,k)-2.*phi(i,k))*dy2i &
                  +(phi(i,k+1)+phi(i,k-1)-2.*phi(i,k))*dx2i
             dphi = dphi*dt
             dphimax = max(dphimax,dphi)
             phin(i,k) = phi(i,k)+dphi
          enddo
       enddo
  
       ! save values
       do k = 1,n-2
          do i =1,m-2
             phi(i,k) =phin(i,k)
          enddo
       enddo
  
       if(dphimax < eps) EXIT
  
    enddo
    if (dphimax >= eps) it=it-1
  
    call cpu_time(t1)
    call w_time(wt1)
    if (wt1<wt0) wt1 = wt1 + 24*3600d0 
 
    ! print array
    write (*,'(/,a,i6,a)') 'phi after',it,' iterations'
  
    call heatpr
  
    write (*,'(/,a,f12.6)')  'cpu_time  :   ', t1-t0
    write (*,'(  a, f9.3)')  'wallclock :   ', wt1-wt0
  
  
  end subroutine heat_sub
 
end module heat_mod
 
 
program heat

  use heat_mod 
  implicit none
 
  integer :: dim1, dim2, argc, arglen
  character(len=20) :: argval

  argc = command_argument_count()
  if (argc < 2) then
      print *,'Synopsis: heat <dim1> <dim2> [<itmax>]'
      print *
      print *,'Examples: heat 30 30'
      print *,'          heat 30 30 15000'
    stop
  end if 
  call get_command_argument(1, argval, arglen)
  read(argval,*) dim1
  call get_command_argument(2, argval, arglen)
  read(argval,*) dim2
  itmax=15000 
  if (argc >= 3) then
    call get_command_argument(3, argval, arglen)
    read(argval,*) itmax
  end if 
 
  m=dim1+2  ! two additional elements for the boundary conditions 
  n=dim2+2
  print '(''main:   dim1  ='',i4,''   dim2  ='',i4,''   itmax=''i6)', dim1, dim2, itmax
  print '(''main: m=dim1+2='',i4,'' n=dim2+2='',i4)', m, n
  
  allocate( phi (0:m-1,0:n-1) ) 
  allocate( phin(1:m-2,1:n-2) )
 
  call heat_sub
 
end program heat
